@@ -22,12 +22,8 @@
 #include "soup-ssl.h"
 
 #include <sys/types.h>
-#include <sys/socket.h>
-#include <netinet/in.h>
-#include <netinet/tcp.h>
 
-#define PARENT_TYPE G_TYPE_OBJECT
-static GObjectClass *parent_class;
+G_DEFINE_TYPE (SoupSocket, soup_socket, G_TYPE_OBJECT)
 
 enum {
 	CONNECT_RESULT,
@@ -46,13 +42,16 @@ enum {
 	PROP_NON_BLOCKING,
 	PROP_NODELAY,
 	PROP_REUSEADDR,
+	PROP_CLOEXEC,
 	PROP_IS_SERVER,
 	PROP_SSL_CREDENTIALS,
+	PROP_ASYNC_CONTEXT,
+	PROP_TIMEOUT,
 
 	LAST_PROP
 };
 
-struct SoupSocketPrivate {
+typedef struct {
 	int sockfd;
 	SoupAddress *local_addr, *remote_addr;
 	GIOChannel *iochannel;
@@ -60,15 +59,19 @@ struct SoupSocketPrivate {
 	guint non_blocking:1;
 	guint nodelay:1;
 	guint reuseaddr:1;
+	guint cloexec:1;
 	guint is_server:1;
 	gpointer ssl_creds;
 
-	guint           watch;
-	guint           read_tag, write_tag, error_tag;
+	GMainContext   *async_context;
+	GSource        *watch_src;
+	GSource        *read_src, *write_src;
 	GByteArray     *read_buf;
 
 	GMutex *iolock, *addrlock;
-};
+	guint timeout;
+} SoupSocketPrivate;
+#define SOUP_SOCKET_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), SOUP_TYPE_SOCKET, SoupSocketPrivate))
 
 #ifdef HAVE_IPV6
 #define soup_sockaddr_max sockaddr_in6
@@ -81,68 +84,82 @@ static void set_property (GObject *object, guint prop_id,
 static void get_property (GObject *object, guint prop_id,
 			  GValue *value, GParamSpec *pspec);
 
-static void
-init (GObject *object)
-{
-	SoupSocket *sock = SOUP_SOCKET (object);
+#ifdef G_OS_WIN32
+#define SOUP_CLOSE_SOCKET(socket) closesocket (socket)
+#define SOUP_IS_SOCKET_ERROR(status) ((status) == SOCKET_ERROR)
+#define SOUP_IS_INVALID_SOCKET(socket) ((socket) == INVALID_SOCKET)
+#define SOUP_IS_CONNECT_STATUS_INPROGRESS() (WSAGetLastError () == WSAEWOULDBLOCK)
+#else
+#define SOUP_CLOSE_SOCKET(socket) close (socket)
+#define SOUP_IS_SOCKET_ERROR(status) ((status) == -1)
+#define SOUP_IS_INVALID_SOCKET(socket) ((socket) < 0)
+#define SOUP_IS_CONNECT_STATUS_INPROGRESS() (errno == EINPROGRESS)
+#endif
 
-	sock->priv = g_new0 (SoupSocketPrivate, 1);
-	sock->priv->sockfd = -1;
-	sock->priv->non_blocking = sock->priv->nodelay = TRUE;
-	sock->priv->reuseaddr = TRUE;
-	sock->priv->addrlock = g_mutex_new ();
-	sock->priv->iolock = g_mutex_new ();
+static void
+soup_socket_init (SoupSocket *sock)
+{
+	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (sock);
+
+	priv->sockfd = -1;
+	priv->non_blocking = priv->nodelay = TRUE;
+	priv->reuseaddr = TRUE;
+	priv->cloexec = FALSE;
+	priv->addrlock = g_mutex_new ();
+	priv->iolock = g_mutex_new ();
+	priv->timeout = 0;
 }
 
 static void
-disconnect_internal (SoupSocket *sock)
+disconnect_internal (SoupSocketPrivate *priv)
 {
-	g_io_channel_unref (sock->priv->iochannel);
-	sock->priv->iochannel = NULL;
-	sock->priv->sockfd = -1;
+	g_io_channel_unref (priv->iochannel);
+	priv->iochannel = NULL;
+	priv->sockfd = -1;
 
-	if (sock->priv->read_tag) {
-		g_source_remove (sock->priv->read_tag);
-		sock->priv->read_tag = 0;
+	if (priv->read_src) {
+		g_source_destroy (priv->read_src);
+		priv->read_src = NULL;
 	}
-	if (sock->priv->write_tag) {
-		g_source_remove (sock->priv->write_tag);
-		sock->priv->write_tag = 0;
-	}
-	if (sock->priv->error_tag) {
-		g_source_remove (sock->priv->error_tag);
-		sock->priv->error_tag = 0;
+	if (priv->write_src) {
+		g_source_destroy (priv->write_src);
+		priv->write_src = NULL;
 	}
 }
 
 static void
 finalize (GObject *object)
 {
-	SoupSocket *sock = SOUP_SOCKET (object);
+	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (object);
 
-	if (sock->priv->iochannel)
-		disconnect_internal (sock);
+	if (priv->iochannel)
+		disconnect_internal (priv);
 
-	if (sock->priv->local_addr)
-		g_object_unref (sock->priv->local_addr);
-	if (sock->priv->remote_addr)
-		g_object_unref (sock->priv->remote_addr);
+	if (priv->local_addr)
+		g_object_unref (priv->local_addr);
+	if (priv->remote_addr)
+		g_object_unref (priv->remote_addr);
 
-	if (sock->priv->watch)
-		g_source_remove (sock->priv->watch);
+	if (priv->watch_src)
+		g_source_destroy (priv->watch_src);
+	if (priv->async_context)
+		g_main_context_unref (priv->async_context);
 
-	g_mutex_free (sock->priv->addrlock);
-	g_mutex_free (sock->priv->iolock);
+	if (priv->read_buf)
+		g_byte_array_free (priv->read_buf, TRUE);
 
-	g_free (sock->priv);
+	g_mutex_free (priv->addrlock);
+	g_mutex_free (priv->iolock);
 
-	G_OBJECT_CLASS (parent_class)->finalize (object);
+	G_OBJECT_CLASS (soup_socket_parent_class)->finalize (object);
 }
 
 static void
-class_init (GObjectClass *object_class)
+soup_socket_class_init (SoupSocketClass *socket_class)
 {
-	parent_class = g_type_class_ref (PARENT_TYPE);
+	GObjectClass *object_class = G_OBJECT_CLASS (socket_class);
+
+	g_type_class_add_private (socket_class, sizeof (SoupSocketPrivate));
 
 	/* virtual method override */
 	object_class->finalize = finalize;
@@ -150,6 +167,15 @@ class_init (GObjectClass *object_class)
 	object_class->get_property = get_property;
 
 	/* signals */
+
+	/**
+	 * SoupSocket::connect-result:
+	 * @sock: the socket
+	 * @status: the status
+	 *
+	 * Emitted when a connection attempt succeeds or fails. This
+	 * is used internally by soup_socket_client_new_async().
+	 **/
 	signals[CONNECT_RESULT] =
 		g_signal_new ("connect_result",
 			      G_OBJECT_CLASS_TYPE (object_class),
@@ -159,6 +185,14 @@ class_init (GObjectClass *object_class)
 			      soup_marshal_NONE__INT,
 			      G_TYPE_NONE, 1,
 			      G_TYPE_INT);
+
+	/**
+	 * SoupSocket::readable:
+	 * @sock: the socket
+	 *
+	 * Emitted when an async socket is readable. See
+	 * soup_socket_read() and soup_socket_read_until().
+	 **/
 	signals[READABLE] =
 		g_signal_new ("readable",
 			      G_OBJECT_CLASS_TYPE (object_class),
@@ -167,6 +201,14 @@ class_init (GObjectClass *object_class)
 			      NULL, NULL,
 			      soup_marshal_NONE__NONE,
 			      G_TYPE_NONE, 0);
+
+	/**
+	 * SoupSocket::writable:
+	 * @sock: the socket
+	 *
+	 * Emitted when an async socket is writable. See
+	 * soup_socket_write().
+	 **/
 	signals[WRITABLE] =
 		g_signal_new ("writable",
 			      G_OBJECT_CLASS_TYPE (object_class),
@@ -175,6 +217,14 @@ class_init (GObjectClass *object_class)
 			      NULL, NULL,
 			      soup_marshal_NONE__NONE,
 			      G_TYPE_NONE, 0);
+
+	/**
+	 * SoupSocket::disconnected:
+	 * @sock: the socket
+	 *
+	 * Emitted when the socket is disconnected, for whatever
+	 * reason.
+	 **/
 	signals[DISCONNECTED] =
 		g_signal_new ("disconnected",
 			      G_OBJECT_CLASS_TYPE (object_class),
@@ -183,6 +233,16 @@ class_init (GObjectClass *object_class)
 			      NULL, NULL,
 			      soup_marshal_NONE__NONE,
 			      G_TYPE_NONE, 0);
+
+	/**
+	 * SoupSocket::new-connection:
+	 * @sock: the socket
+	 * @new: the new socket
+	 *
+	 * Emitted when a listening socket (set up with
+	 * soup_socket_listen() or soup_socket_server_new()) receives a
+	 * new connection.
+	 **/
 	signals[NEW_CONNECTION] =
 		g_signal_new ("new_connection",
 			      G_OBJECT_CLASS_TYPE (object_class),
@@ -215,6 +275,13 @@ class_init (GObjectClass *object_class)
 				      "Whether or not the socket uses the TCP REUSEADDR flag",
 				      TRUE,
 				      G_PARAM_READWRITE));
+	g_object_class_install_property (
+		object_class, PROP_CLOEXEC,
+		g_param_spec_boolean (SOUP_SOCKET_FLAG_CLOEXEC,
+				      "CLOEXEC",
+				      "Whether or not the socket will be closed automatically on exec()",
+				      FALSE,
+				      G_PARAM_READWRITE));
 	g_object_class_install_property (
 		object_class, PROP_IS_SERVER,
 		g_param_spec_boolean (SOUP_SOCKET_IS_SERVER,
@@ -228,58 +295,120 @@ class_init (GObjectClass *object_class)
 				      "SSL credentials",
 				      "SSL credential information, passed from the session to the SSL implementation",
 				      G_PARAM_READWRITE));
-}
+	g_object_class_install_property (
+		object_class, PROP_ASYNC_CONTEXT,
+		g_param_spec_pointer (SOUP_SOCKET_ASYNC_CONTEXT,
+				      "Async GMainContext",
+				      "The GMainContext to dispatch this socket's async I/O in",
+				      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 
-SOUP_MAKE_TYPE (soup_socket, SoupSocket, class_init, init, PARENT_TYPE)
+	g_object_class_install_property (
+		object_class, PROP_TIMEOUT,
+		g_param_spec_uint (SOUP_SOCKET_TIMEOUT,
+				   "Timeout value",
+				   "Value in seconds to timeout a blocking I/O",
+				   0, G_MAXUINT, 0,
+				   G_PARAM_READWRITE));
+
+#ifdef G_OS_WIN32
+	/* Make sure WSAStartup() gets called. */
+	soup_address_get_type ();
+#endif
+}
 
 
 static void
-update_fdflags (SoupSocket *sock)
+update_fdflags (SoupSocketPrivate *priv)
 {
-	int flags, opt;
+	int opt;
+	struct timeval timeout;
+#ifndef G_OS_WIN32
+	int flags;
+#endif
 
-	if (sock->priv->sockfd == -1)
+	if (priv->sockfd == -1)
 		return;
 
-	flags = fcntl (sock->priv->sockfd, F_GETFL, 0);
+#ifndef G_OS_WIN32
+	flags = fcntl (priv->sockfd, F_GETFL, 0);
 	if (flags != -1) {
-		if (sock->priv->non_blocking)
+		if (priv->non_blocking)
 			flags |= O_NONBLOCK;
 		else
 			flags &= ~O_NONBLOCK;
-		fcntl (sock->priv->sockfd, F_SETFL, flags);
+		fcntl (priv->sockfd, F_SETFL, flags);
 	}
+       flags = fcntl (priv->sockfd, F_GETFD, 0);
+       if (flags != -1) {
+               if (priv->cloexec)
+                       flags |= FD_CLOEXEC;
+               else
+                       flags &= ~FD_CLOEXEC;
+               fcntl (priv->sockfd, F_SETFD, flags);
+        }
 
-	opt = (sock->priv->nodelay != 0);
-	setsockopt (sock->priv->sockfd, IPPROTO_TCP,
-		    TCP_NODELAY, &opt, sizeof (opt));
+#else
+	if (priv->non_blocking) {
+		u_long val = 1;
+		ioctlsocket (priv->sockfd, FIONBIO, &val);
+	} else {
+		u_long val = 0;
+		ioctlsocket (priv->sockfd, FIONBIO, &val);
+	}		
+#endif
 
-	opt = (sock->priv->reuseaddr != 0);
-	setsockopt (sock->priv->sockfd, SOL_SOCKET,
-		    SO_REUSEADDR, &opt, sizeof (opt));
+	opt = (priv->nodelay != 0);
+	setsockopt (priv->sockfd, IPPROTO_TCP,
+		    TCP_NODELAY, (void *) &opt, sizeof (opt));
+
+	timeout.tv_sec = priv->timeout;
+	timeout.tv_usec = 0;
+	setsockopt (priv->sockfd, SOL_SOCKET,
+		    SO_RCVTIMEO, (void *) &timeout, sizeof (timeout));
+
+	timeout.tv_sec = priv->timeout;
+	timeout.tv_usec = 0;
+	setsockopt (priv->sockfd, SOL_SOCKET,
+		    SO_SNDTIMEO, (void *) &timeout, sizeof (timeout));
+
+	opt = (priv->reuseaddr != 0);
+	setsockopt (priv->sockfd, SOL_SOCKET,
+		    SO_REUSEADDR, (void *) &opt, sizeof (opt));
 }
 
 static void
 set_property (GObject *object, guint prop_id,
 	      const GValue *value, GParamSpec *pspec)
 {
-	SoupSocket *sock = SOUP_SOCKET (object);
+	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (object);
 
 	switch (prop_id) {
 	case PROP_NON_BLOCKING:
-		sock->priv->non_blocking = g_value_get_boolean (value);
-		update_fdflags (sock);
+		priv->non_blocking = g_value_get_boolean (value);
+		update_fdflags (priv);
 		break;
 	case PROP_NODELAY:
-		sock->priv->nodelay = g_value_get_boolean (value);
-		update_fdflags (sock);
+		priv->nodelay = g_value_get_boolean (value);
+		update_fdflags (priv);
 		break;
 	case PROP_REUSEADDR:
-		sock->priv->reuseaddr = g_value_get_boolean (value);
-		update_fdflags (sock);
+		priv->reuseaddr = g_value_get_boolean (value);
+		update_fdflags (priv);
+		break;
+	case PROP_CLOEXEC:
+		priv->cloexec = g_value_get_boolean (value);
+		update_fdflags (priv);
 		break;
 	case PROP_SSL_CREDENTIALS:
-		sock->priv->ssl_creds = g_value_get_pointer (value);
+		priv->ssl_creds = g_value_get_pointer (value);
+		break;
+	case PROP_ASYNC_CONTEXT:
+		priv->async_context = g_value_get_pointer (value);
+		if (priv->async_context)
+			g_main_context_ref (priv->async_context);
+		break;
+	case PROP_TIMEOUT:
+		priv->timeout = g_value_get_uint (value);
 		break;
 	default:
 		break;
@@ -290,23 +419,32 @@ static void
 get_property (GObject *object, guint prop_id,
 	      GValue *value, GParamSpec *pspec)
 {
-	SoupSocket *sock = SOUP_SOCKET (object);
+	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (object);
 
 	switch (prop_id) {
 	case PROP_NON_BLOCKING:
-		g_value_set_boolean (value, sock->priv->non_blocking);
+		g_value_set_boolean (value, priv->non_blocking);
 		break;
 	case PROP_NODELAY:
-		g_value_set_boolean (value, sock->priv->nodelay);
+		g_value_set_boolean (value, priv->nodelay);
 		break;
 	case PROP_REUSEADDR:
-		g_value_set_boolean (value, sock->priv->reuseaddr);
+		g_value_set_boolean (value, priv->reuseaddr);
 		break;
+	case PROP_CLOEXEC:
+		g_value_set_boolean (value, priv->cloexec);
+                break;
 	case PROP_IS_SERVER:
-		g_value_set_boolean (value, sock->priv->is_server);
+		g_value_set_boolean (value, priv->is_server);
 		break;
 	case PROP_SSL_CREDENTIALS:
-		g_value_set_pointer (value, sock->priv->ssl_creds);
+		g_value_set_pointer (value, priv->ssl_creds);
+		break;
+	case PROP_ASYNC_CONTEXT:
+		g_value_set_pointer (value, priv->async_context ? g_main_context_ref (priv->async_context) : NULL);
+		break;
+	case PROP_TIMEOUT:
+		g_value_set_uint (value, priv->timeout);
 		break;
 	default:
 		break;
@@ -338,29 +476,35 @@ soup_socket_new (const char *optname1, ...)
 }
 
 static GIOChannel *
-get_iochannel (SoupSocket *sock)
+get_iochannel (SoupSocketPrivate *priv)
 {
-	g_mutex_lock (sock->priv->iolock);
-	if (!sock->priv->iochannel) {
-		sock->priv->iochannel =
-			g_io_channel_unix_new (sock->priv->sockfd);
-		g_io_channel_set_close_on_unref (sock->priv->iochannel, TRUE);
-		g_io_channel_set_encoding (sock->priv->iochannel, NULL, NULL);
-		g_io_channel_set_buffered (sock->priv->iochannel, FALSE);
+	g_mutex_lock (priv->iolock);
+	if (!priv->iochannel) {
+#ifndef G_OS_WIN32
+		priv->iochannel =
+			g_io_channel_unix_new (priv->sockfd);
+#else
+		priv->iochannel =
+			g_io_channel_win32_new_socket (priv->sockfd);
+#endif
+		g_io_channel_set_close_on_unref (priv->iochannel, TRUE);
+		g_io_channel_set_encoding (priv->iochannel, NULL, NULL);
+		g_io_channel_set_buffered (priv->iochannel, FALSE);
 	}
-	g_mutex_unlock (sock->priv->iolock);
-	return sock->priv->iochannel;
+	g_mutex_unlock (priv->iolock);
+	return priv->iochannel;
 }
 
 static gboolean
 idle_connect_result (gpointer user_data)
 {
 	SoupSocket *sock = user_data;
+	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (sock);
 
-	sock->priv->watch = 0;
+	priv->watch_src = NULL;
 
 	g_signal_emit (sock, signals[CONNECT_RESULT], 0,
-		       sock->priv->sockfd != -1 ? SOUP_STATUS_OK : SOUP_STATUS_CANT_CONNECT);
+		       priv->sockfd != -1 ? SOUP_STATUS_OK : SOUP_STATUS_CANT_CONNECT);
 	return FALSE;
 }
 
@@ -368,18 +512,19 @@ static gboolean
 connect_watch (GIOChannel* iochannel, GIOCondition condition, gpointer data)
 {
 	SoupSocket *sock = data;
+	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (sock);
 	int error = 0;
 	int len = sizeof (error);
 
 	/* Remove the watch now in case we don't return immediately */
-	g_source_remove (sock->priv->watch);
-	sock->priv->watch = 0;
+	g_source_destroy (priv->watch_src);
+	priv->watch_src = NULL;
 
 	if (condition & ~(G_IO_IN | G_IO_OUT))
 		goto cant_connect;
 
-	if (getsockopt (sock->priv->sockfd, SOL_SOCKET, SO_ERROR,
-			&error, &len) != 0)
+	if (getsockopt (priv->sockfd, SOL_SOCKET, SO_ERROR,
+			(void *) &error, &len) != 0)
 		goto cant_connect;
 	if (error)
 		goto cant_connect;
@@ -395,6 +540,7 @@ static void
 got_address (SoupAddress *addr, guint status, gpointer user_data)
 {
 	SoupSocket *sock = user_data;
+	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (sock);
 
 	if (!SOUP_STATUS_IS_SUCCESSFUL (status)) {
 		g_signal_emit (sock, signals[CONNECT_RESULT], 0, status);
@@ -402,7 +548,7 @@ got_address (SoupAddress *addr, guint status, gpointer user_data)
 		return;
 	}
 
-	soup_socket_connect (sock, sock->priv->remote_addr);
+	soup_socket_connect (sock, priv->remote_addr);
 	/* soup_socket_connect re-reffed addr */
 	g_object_unref (addr);
 
@@ -428,102 +574,110 @@ got_address (SoupAddress *addr, guint status, gpointer user_data)
 guint
 soup_socket_connect (SoupSocket *sock, SoupAddress *remote_addr)
 {
+	SoupSocketPrivate *priv;
 	struct sockaddr *sa;
 	int len, status;
 
 	g_return_val_if_fail (SOUP_IS_SOCKET (sock), SOUP_STATUS_MALFORMED);
-	g_return_val_if_fail (!sock->priv->is_server, SOUP_STATUS_MALFORMED);
-	g_return_val_if_fail (sock->priv->sockfd == -1, SOUP_STATUS_MALFORMED);
+	priv = SOUP_SOCKET_GET_PRIVATE (sock);
+	g_return_val_if_fail (!priv->is_server, SOUP_STATUS_MALFORMED);
+	g_return_val_if_fail (priv->sockfd == -1, SOUP_STATUS_MALFORMED);
 	g_return_val_if_fail (SOUP_IS_ADDRESS (remote_addr), SOUP_STATUS_MALFORMED);
 
-	sock->priv->remote_addr = g_object_ref (remote_addr);
-	if (!sock->priv->non_blocking) {
+	priv->remote_addr = g_object_ref (remote_addr);
+	if (!priv->non_blocking) {
 		status = soup_address_resolve_sync (remote_addr);
 		if (!SOUP_STATUS_IS_SUCCESSFUL (status))
 			return status;
 	}
 
-	sa = soup_address_get_sockaddr (sock->priv->remote_addr, &len);
+	sa = soup_address_get_sockaddr (priv->remote_addr, &len);
 	if (!sa) {
-		if (!sock->priv->non_blocking)
+		if (!priv->non_blocking)
 			return SOUP_STATUS_CANT_RESOLVE;
 
 		g_object_ref (sock);
-		soup_address_resolve_async (remote_addr, got_address, sock);
+		soup_address_resolve_async_full (remote_addr, priv->async_context,
+						 got_address, sock);
 		return SOUP_STATUS_CONTINUE;
 	}
 
-	sock->priv->sockfd = socket (sa->sa_family, SOCK_STREAM, 0);
-	if (sock->priv->sockfd == -1) {
+	priv->sockfd = socket (sa->sa_family, SOCK_STREAM, 0);
+	if (SOUP_IS_INVALID_SOCKET (priv->sockfd)) {
 		goto done;
 	}
-	update_fdflags (sock);
+	update_fdflags (priv);
 
-	status = connect (sock->priv->sockfd, sa, len);
+	status = connect (priv->sockfd, sa, len);
 
-	if (status == -1) {
-		if (errno == EINPROGRESS) {
+	if (SOUP_IS_SOCKET_ERROR (status)) {
+		if (SOUP_IS_CONNECT_STATUS_INPROGRESS ()) {
 			/* Wait for connect to succeed or fail */
-			sock->priv->watch =
-				g_io_add_watch (get_iochannel (sock),
-						G_IO_IN | G_IO_OUT |
-						G_IO_PRI | G_IO_ERR |
-						G_IO_HUP | G_IO_NVAL,
-						connect_watch, sock);
+			priv->watch_src =
+				soup_add_io_watch (priv->async_context,
+						   get_iochannel (priv),
+						   G_IO_IN | G_IO_OUT |
+						   G_IO_PRI | G_IO_ERR |
+						   G_IO_HUP | G_IO_NVAL,
+						   connect_watch, sock);
 			return SOUP_STATUS_CONTINUE;
 		} else {
-			close (sock->priv->sockfd);
-			sock->priv->sockfd = -1;
+			SOUP_CLOSE_SOCKET (priv->sockfd);
+			priv->sockfd = -1;
 		}
-	}
+	} else
+		get_iochannel (priv);
 
  done:
-	if (sock->priv->non_blocking) {
-		sock->priv->watch = g_idle_add (idle_connect_result, sock);
+	if (priv->non_blocking) {
+		priv->watch_src = soup_add_idle (priv->async_context,
+						 idle_connect_result, sock);
 		return SOUP_STATUS_CONTINUE;
-	} else if (sock->priv->sockfd == -1)
+	} else if (SOUP_IS_INVALID_SOCKET (priv->sockfd))
 		return SOUP_STATUS_CANT_CONNECT;
-	else {
-		get_iochannel (sock);
+	else
 		return SOUP_STATUS_OK;
-	}
 }
 
 static gboolean
 listen_watch (GIOChannel* iochannel, GIOCondition condition, gpointer data)
 {
 	SoupSocket *sock = data, *new;
+	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (sock), *new_priv;
 	struct soup_sockaddr_max sa;
 	int sa_len, sockfd;
 
 	if (condition & (G_IO_HUP | G_IO_ERR)) {
-		g_source_remove (sock->priv->watch);
-		sock->priv->watch = 0;
+		g_source_destroy (priv->watch_src);
+		priv->watch_src = NULL;
 		return FALSE;
 	}
 
 	sa_len = sizeof (sa);
-	sockfd = accept (sock->priv->sockfd, (struct sockaddr *)&sa, &sa_len);
-	if (sockfd == -1)
+	sockfd = accept (priv->sockfd, (struct sockaddr *)&sa, &sa_len);
+	if (SOUP_IS_INVALID_SOCKET (sockfd))
 		return TRUE;
 
 	new = g_object_new (SOUP_TYPE_SOCKET, NULL);
-	new->priv->sockfd = sockfd;
-	new->priv->non_blocking = sock->priv->non_blocking;
-	new->priv->nodelay = sock->priv->nodelay;
-	new->priv->is_server = TRUE;
-	new->priv->ssl_creds = sock->priv->ssl_creds;
-	update_fdflags (new);
+	new_priv = SOUP_SOCKET_GET_PRIVATE (new);
+	new_priv->sockfd = sockfd;
+	if (priv->async_context)
+		new_priv->async_context = g_main_context_ref (priv->async_context);
+	new_priv->non_blocking = priv->non_blocking;
+	new_priv->nodelay = priv->nodelay;
+	new_priv->is_server = TRUE;
+	new_priv->ssl_creds = priv->ssl_creds;
+	update_fdflags (new_priv);
 
-	new->priv->remote_addr = soup_address_new_from_sockaddr ((struct sockaddr *)&sa, sa_len);
+	new_priv->remote_addr = soup_address_new_from_sockaddr ((struct sockaddr *)&sa, sa_len);
 
-	if (new->priv->ssl_creds) {
+	if (new_priv->ssl_creds) {
 		if (!soup_socket_start_ssl (new)) {
 			g_object_unref (new);
 			return TRUE;
 		}
 	} else
-		get_iochannel (new);
+		get_iochannel (new_priv);
 
 	g_signal_emit (sock, signals[NEW_CONNECTION], 0, new);
 	g_object_unref (new);
@@ -545,16 +699,19 @@ listen_watch (GIOChannel* iochannel, GIOCondition condition, gpointer data)
 gboolean
 soup_socket_listen (SoupSocket *sock, SoupAddress *local_addr)
 {
+	SoupSocketPrivate *priv;
 	struct sockaddr *sa;
 	int sa_len;
 
 	g_return_val_if_fail (SOUP_IS_SOCKET (sock), FALSE);
-	g_return_val_if_fail (sock->priv->is_server, FALSE);
-	g_return_val_if_fail (sock->priv->sockfd == -1, FALSE);
+	priv = SOUP_SOCKET_GET_PRIVATE (sock);
+	g_return_val_if_fail (priv->sockfd == -1, FALSE);
 	g_return_val_if_fail (SOUP_IS_ADDRESS (local_addr), FALSE);
 
+	priv->is_server = TRUE;
+
 	/* @local_addr may have its port set to 0. So we intentionally
-	 * don't store it in sock->priv->local_addr, so that if the
+	 * don't store it in priv->local_addr, so that if the
 	 * caller calls soup_socket_get_local_address() later, we'll
 	 * have to make a new addr by calling getsockname(), which
 	 * will have the right port number.
@@ -562,28 +719,29 @@ soup_socket_listen (SoupSocket *sock, SoupAddress *local_addr)
 	sa = soup_address_get_sockaddr (local_addr, &sa_len);
 	g_return_val_if_fail (sa != NULL, FALSE);
 
-	sock->priv->sockfd = socket (sa->sa_family, SOCK_STREAM, 0);
-	if (sock->priv->sockfd < 0)
+	priv->sockfd = socket (sa->sa_family, SOCK_STREAM, 0);
+	if (SOUP_IS_INVALID_SOCKET (priv->sockfd))
 		goto cant_listen;
-	update_fdflags (sock);
+	update_fdflags (priv);
 
 	/* Bind */
-	if (bind (sock->priv->sockfd, sa, sa_len) != 0)
+	if (bind (priv->sockfd, sa, sa_len) != 0)
 		goto cant_listen;
 
 	/* Listen */
-	if (listen (sock->priv->sockfd, 10) != 0)
+	if (listen (priv->sockfd, 10) != 0)
 		goto cant_listen;
 
-	sock->priv->watch = g_io_add_watch (get_iochannel (sock),
-					    G_IO_IN | G_IO_ERR | G_IO_HUP,
-					    listen_watch, sock);
+	priv->watch_src = soup_add_io_watch (priv->async_context,
+					     get_iochannel (priv),
+					     G_IO_IN | G_IO_ERR | G_IO_HUP,
+					     listen_watch, sock);
 	return TRUE;
 
  cant_listen:
-	if (sock->priv->sockfd != -1) {
-		close (sock->priv->sockfd);
-		sock->priv->sockfd = -1;
+	if (priv->sockfd != -1) {
+		SOUP_CLOSE_SOCKET (priv->sockfd);
+		priv->sockfd = -1;
 	}
 
 	return FALSE;
@@ -600,19 +758,37 @@ soup_socket_listen (SoupSocket *sock, SoupAddress *local_addr)
 gboolean
 soup_socket_start_ssl (SoupSocket *sock)
 {
+	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (sock);
+
+	return soup_socket_start_proxy_ssl (sock, soup_address_get_name (priv->remote_addr));
+}
+	
+/**
+ * soup_socket_start_proxy_ssl:
+ * @sock: the socket
+ * @ssl_host: hostname of the SSL server
+ *
+ * Starts using SSL on @socket, expecting to find a host named
+ * @ssl_host.
+ *
+ * Return value: success or failure
+ **/
+gboolean
+soup_socket_start_proxy_ssl (SoupSocket *sock, const char *ssl_host)
+{
+	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (sock);
 	GIOChannel *ssl_chan;
 
-	get_iochannel (sock);
+	get_iochannel (priv);
 	ssl_chan = soup_ssl_wrap_iochannel (
-		sock->priv->iochannel, sock->priv->is_server ?
+		priv->iochannel, priv->is_server ?
 		SOUP_SSL_TYPE_SERVER : SOUP_SSL_TYPE_CLIENT,
-		soup_address_get_name (sock->priv->remote_addr),
-		sock->priv->ssl_creds);
+		ssl_host, priv->ssl_creds);
 
 	if (!ssl_chan)
 		return FALSE;
 
-	sock->priv->iochannel = ssl_chan;
+	priv->iochannel = ssl_chan;
 	return TRUE;
 }
 	
@@ -628,6 +804,9 @@ soup_socket_start_ssl (SoupSocket *sock)
  * Creates a connection to @hostname and @port. @callback will be
  * called when the connection completes (or fails).
  *
+ * Uses the default #GMainContext. If you need to use an alternate
+ * context, use soup_socket_new() and soup_socket_connect() directly.
+ *
  * Return value: the new socket (not yet ready for use).
  **/
 SoupSocket *
@@ -636,13 +815,16 @@ soup_socket_client_new_async (const char *hostname, guint port,
 			      SoupSocketCallback callback, gpointer user_data)
 {
 	SoupSocket *sock;
+	SoupAddress *addr;
 
 	g_return_val_if_fail (hostname != NULL, NULL);
 
 	sock = g_object_new (SOUP_TYPE_SOCKET,
 			     SOUP_SOCKET_SSL_CREDENTIALS, ssl_creds,
 			     NULL);
-	soup_socket_connect (sock, soup_address_new (hostname, port));
+	addr = soup_address_new (hostname, port);
+	soup_socket_connect (sock, addr);
+	g_object_unref (addr);
 
 	if (callback) {
 		soup_signal_connect_once (sock, "connect_result",
@@ -668,6 +850,8 @@ soup_socket_client_new_sync (const char *hostname, guint port,
 			     gpointer ssl_creds, guint *status_ret)
 {
 	SoupSocket *sock;
+	SoupSocketPrivate *priv;
+	SoupAddress *addr;
 	guint status;
 
 	g_return_val_if_fail (hostname != NULL, NULL);
@@ -675,8 +859,11 @@ soup_socket_client_new_sync (const char *hostname, guint port,
 	sock = g_object_new (SOUP_TYPE_SOCKET,
 			     SOUP_SOCKET_SSL_CREDENTIALS, ssl_creds,
 			     NULL);
-	sock->priv->non_blocking = FALSE;
-	status = soup_socket_connect (sock, soup_address_new (hostname, port));
+	priv = SOUP_SOCKET_GET_PRIVATE (sock);
+	priv->non_blocking = FALSE;
+	addr = soup_address_new (hostname, port);
+	status = soup_socket_connect (sock, addr);
+	g_object_unref (addr);
 
 	if (!SOUP_STATUS_IS_SUCCESSFUL (status)) {
 		g_object_unref (sock);
@@ -700,6 +887,9 @@ soup_socket_client_new_sync (const char *hostname, guint port,
  * address. @callback will be called each time a client connects,
  * with a new #SoupSocket.
  *
+ * Uses the default #GMainContext. If you need to use an alternate
+ * context, use soup_socket_new() and soup_socket_listen() directly.
+ *
  * Returns: a new #SoupSocket, or NULL if there was a failure.
  **/
 SoupSocket *
@@ -708,13 +898,14 @@ soup_socket_server_new (SoupAddress *local_addr, gpointer ssl_creds,
 			gpointer user_data)
 {
 	SoupSocket *sock;
+	SoupSocketPrivate *priv;
 
 	g_return_val_if_fail (SOUP_IS_ADDRESS (local_addr), NULL);
 
 	sock = g_object_new (SOUP_TYPE_SOCKET,
 			     SOUP_SOCKET_SSL_CREDENTIALS, ssl_creds,
 			     NULL);
-	sock->priv->is_server = TRUE;
+	priv = SOUP_SOCKET_GET_PRIVATE (sock);
 	if (!soup_socket_listen (sock, local_addr)) {
 		g_object_unref (sock);
 		return NULL;
@@ -739,16 +930,18 @@ soup_socket_server_new (SoupAddress *local_addr, gpointer ssl_creds,
 void
 soup_socket_disconnect (SoupSocket *sock)
 {
+	SoupSocketPrivate *priv;
 	gboolean already_disconnected = FALSE;
 
 	g_return_if_fail (SOUP_IS_SOCKET (sock));
+	priv = SOUP_SOCKET_GET_PRIVATE (sock);
 
-	if (g_mutex_trylock (sock->priv->iolock)) {
-		if (sock->priv->iochannel)
-			disconnect_internal (sock);
+	if (g_mutex_trylock (priv->iolock)) {
+		if (priv->iochannel)
+			disconnect_internal (priv);
 		else
 			already_disconnected = TRUE;
-		g_mutex_unlock (sock->priv->iolock);
+		g_mutex_unlock (priv->iolock);
 	} else {
 		int sockfd;
 
@@ -757,14 +950,14 @@ soup_socket_disconnect (SoupSocket *sock)
 		 * the file descriptor out from under it.
 		 */
 
-		sockfd = sock->priv->sockfd;
-		sock->priv->sockfd = -1;
+		sockfd = priv->sockfd;
+		priv->sockfd = -1;
 		if (sockfd == -1)
 			already_disconnected = TRUE;
 		else {
-			g_io_channel_set_close_on_unref (sock->priv->iochannel,
+			g_io_channel_set_close_on_unref (priv->iochannel,
 							 FALSE);
-			close (sockfd);
+			SOUP_CLOSE_SOCKET (sockfd);
 		}
 	}
 
@@ -791,9 +984,12 @@ soup_socket_disconnect (SoupSocket *sock)
 gboolean
 soup_socket_is_connected (SoupSocket *sock)
 {
-	g_return_val_if_fail (SOUP_IS_SOCKET (sock), FALSE);
+	SoupSocketPrivate *priv;
 
-	return sock->priv->iochannel != NULL;
+	g_return_val_if_fail (SOUP_IS_SOCKET (sock), FALSE);
+	priv = SOUP_SOCKET_GET_PRIVATE (sock);
+
+	return priv->iochannel != NULL;
 }
 
 /**
@@ -807,20 +1003,23 @@ soup_socket_is_connected (SoupSocket *sock)
 SoupAddress *
 soup_socket_get_local_address (SoupSocket *sock)
 {
-	g_return_val_if_fail (SOUP_IS_SOCKET (sock), NULL);
+	SoupSocketPrivate *priv;
 
-	g_mutex_lock (sock->priv->addrlock);
-	if (!sock->priv->local_addr) {
+	g_return_val_if_fail (SOUP_IS_SOCKET (sock), NULL);
+	priv = SOUP_SOCKET_GET_PRIVATE (sock);
+
+	g_mutex_lock (priv->addrlock);
+	if (!priv->local_addr) {
 		struct soup_sockaddr_max bound_sa;
 		int sa_len;
 
 		sa_len = sizeof (bound_sa);
-		getsockname (sock->priv->sockfd, (struct sockaddr *)&bound_sa, &sa_len);
-		sock->priv->local_addr = soup_address_new_from_sockaddr ((struct sockaddr *)&bound_sa, sa_len);
+		getsockname (priv->sockfd, (struct sockaddr *)&bound_sa, &sa_len);
+		priv->local_addr = soup_address_new_from_sockaddr ((struct sockaddr *)&bound_sa, sa_len);
 	}
-	g_mutex_unlock (sock->priv->addrlock);
+	g_mutex_unlock (priv->addrlock);
 
-	return sock->priv->local_addr;
+	return priv->local_addr;
 }
 
 /**
@@ -834,20 +1033,23 @@ soup_socket_get_local_address (SoupSocket *sock)
 SoupAddress *
 soup_socket_get_remote_address (SoupSocket *sock)
 {
-	g_return_val_if_fail (SOUP_IS_SOCKET (sock), NULL);
+	SoupSocketPrivate *priv;
 
-	g_mutex_lock (sock->priv->addrlock);
-	if (!sock->priv->remote_addr) {
+	g_return_val_if_fail (SOUP_IS_SOCKET (sock), NULL);
+	priv = SOUP_SOCKET_GET_PRIVATE (sock);
+
+	g_mutex_lock (priv->addrlock);
+	if (!priv->remote_addr) {
 		struct soup_sockaddr_max bound_sa;
 		int sa_len;
 
 		sa_len = sizeof (bound_sa);
-		getpeername (sock->priv->sockfd, (struct sockaddr *)&bound_sa, &sa_len);
-		sock->priv->remote_addr = soup_address_new_from_sockaddr ((struct sockaddr *)&bound_sa, sa_len);
+		getpeername (priv->sockfd, (struct sockaddr *)&bound_sa, &sa_len);
+		priv->remote_addr = soup_address_new_from_sockaddr ((struct sockaddr *)&bound_sa, sa_len);
 	}
-	g_mutex_unlock (sock->priv->addrlock);
+	g_mutex_unlock (priv->addrlock);
 
-	return sock->priv->remote_addr;
+	return priv->remote_addr;
 }
 
 
@@ -857,9 +1059,14 @@ static gboolean
 socket_read_watch (GIOChannel *chan, GIOCondition cond, gpointer user_data)
 {
 	SoupSocket *sock = user_data;
+	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (sock);
 
-	sock->priv->read_tag = 0;
-	g_signal_emit (sock, signals[READABLE], 0);
+	priv->read_src = NULL;
+
+	if (cond & (G_IO_ERR | G_IO_HUP))
+		soup_socket_disconnect (sock);
+	else
+		g_signal_emit (sock, signals[READABLE], 0);
 
 	return FALSE;
 }
@@ -867,14 +1074,15 @@ socket_read_watch (GIOChannel *chan, GIOCondition cond, gpointer user_data)
 static SoupSocketIOStatus
 read_from_network (SoupSocket *sock, gpointer buffer, gsize len, gsize *nread)
 {
+	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (sock);
 	GIOStatus status;
 	GIOCondition cond = G_IO_IN;
 	GError *err = NULL;
 
-	if (!sock->priv->iochannel) 
+	if (!priv->iochannel) 
 		return SOUP_SOCKET_EOF;
 
-	status = g_io_channel_read_chars (sock->priv->iochannel,
+	status = g_io_channel_read_chars (priv->iochannel,
 					  buffer, len, nread, &err);
 	if (err) {
 		if (err->domain == SOUP_SSL_ERROR &&
@@ -895,10 +1103,19 @@ read_from_network (SoupSocket *sock, gpointer buffer, gsize len, gsize *nread)
 		if (*nread > 0)
 			return SOUP_SOCKET_OK;
 
-		if (!sock->priv->read_tag) {
-			sock->priv->read_tag =
-				g_io_add_watch (sock->priv->iochannel, cond,
-						socket_read_watch, sock);
+		/* If the connection/session is sync and  we get
+		   EAGAIN or EWOULDBLOCK, then it will be socket timeout
+		   and should be treated as an error condition.
+		*/
+		if (!priv->non_blocking)
+			return SOUP_SOCKET_ERROR;
+
+		if (!priv->read_src) {
+			priv->read_src =
+				soup_add_io_watch (priv->async_context,
+						   priv->iochannel,
+						   cond | G_IO_HUP | G_IO_ERR,
+						   socket_read_watch, sock);
 		}
 		return SOUP_SOCKET_WOULD_BLOCK;
 
@@ -913,14 +1130,15 @@ read_from_network (SoupSocket *sock, gpointer buffer, gsize len, gsize *nread)
 static SoupSocketIOStatus
 read_from_buf (SoupSocket *sock, gpointer buffer, gsize len, gsize *nread)
 {
-	GByteArray *read_buf = sock->priv->read_buf;
+	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (sock);
+	GByteArray *read_buf = priv->read_buf;
 
 	*nread = MIN (read_buf->len, len);
 	memcpy (buffer, read_buf->data, *nread);
 
 	if (*nread == read_buf->len) {
 		g_byte_array_free (read_buf, TRUE);
-		sock->priv->read_buf = NULL;
+		priv->read_buf = NULL;
 	} else {
 		memmove (read_buf->data, read_buf->data + *nread, 
 			 read_buf->len - *nread);
@@ -956,16 +1174,18 @@ read_from_buf (SoupSocket *sock, gpointer buffer, gsize len, gsize *nread)
 SoupSocketIOStatus
 soup_socket_read (SoupSocket *sock, gpointer buffer, gsize len, gsize *nread)
 {
+	SoupSocketPrivate *priv;
 	SoupSocketIOStatus status;
 
 	g_return_val_if_fail (SOUP_IS_SOCKET (sock), SOUP_SOCKET_ERROR);
+	priv = SOUP_SOCKET_GET_PRIVATE (sock);
 
-	g_mutex_lock (sock->priv->iolock);
-	if (sock->priv->read_buf)
+	g_mutex_lock (priv->iolock);
+	if (priv->read_buf)
 		status = read_from_buf (sock, buffer, len, nread);
 	else
 		status = read_from_network (sock, buffer, len, nread);
-	g_mutex_unlock (sock->priv->iolock);
+	g_mutex_unlock (priv->iolock);
 
 	return status;
 }
@@ -993,21 +1213,23 @@ soup_socket_read_until (SoupSocket *sock, gpointer buffer, gsize len,
 			gconstpointer boundary, gsize boundary_len,
 			gsize *nread, gboolean *got_boundary)
 {
+	SoupSocketPrivate *priv;
 	SoupSocketIOStatus status;
 	GByteArray *read_buf;
 	guint match_len, prev_len;
 	guint8 *p, *end;
 
 	g_return_val_if_fail (SOUP_IS_SOCKET (sock), SOUP_SOCKET_ERROR);
+	priv = SOUP_SOCKET_GET_PRIVATE (sock);
 	g_return_val_if_fail (len >= boundary_len, SOUP_SOCKET_ERROR);
 
-	g_mutex_lock (sock->priv->iolock);
+	g_mutex_lock (priv->iolock);
 
 	*got_boundary = FALSE;
 
-	if (!sock->priv->read_buf)
-		sock->priv->read_buf = g_byte_array_new ();
-	read_buf = sock->priv->read_buf;
+	if (!priv->read_buf)
+		priv->read_buf = g_byte_array_new ();
+	read_buf = priv->read_buf;
 
 	if (read_buf->len < boundary_len) {
 		prev_len = read_buf->len;
@@ -1018,7 +1240,7 @@ soup_socket_read_until (SoupSocket *sock, gpointer buffer, gsize len,
 		read_buf->len = prev_len + *nread;
 
 		if (status != SOUP_SOCKET_OK) {
-			g_mutex_unlock (sock->priv->iolock);
+			g_mutex_unlock (priv->iolock);
 			return status;
 		}
 	}
@@ -1040,17 +1262,22 @@ soup_socket_read_until (SoupSocket *sock, gpointer buffer, gsize len,
 	match_len = p - read_buf->data;
 	status = read_from_buf (sock, buffer, MIN (len, match_len), nread);
 
-	g_mutex_unlock (sock->priv->iolock);
+	g_mutex_unlock (priv->iolock);
 	return status;
 }
 
 static gboolean
-socket_write_watch (GIOChannel *chan, GIOCondition condition, gpointer user_data)
+socket_write_watch (GIOChannel *chan, GIOCondition cond, gpointer user_data)
 {
 	SoupSocket *sock = user_data;
+	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (sock);
 
-	sock->priv->write_tag = 0;
-	g_signal_emit (sock, signals[WRITABLE], 0);
+	priv->write_src = NULL;
+
+	if (cond & (G_IO_ERR | G_IO_HUP))
+		soup_socket_disconnect (sock);
+	else
+		g_signal_emit (sock, signals[WRITABLE], 0);
 
 	return FALSE;
 }
@@ -1080,28 +1307,36 @@ SoupSocketIOStatus
 soup_socket_write (SoupSocket *sock, gconstpointer buffer,
 		   gsize len, gsize *nwrote)
 {
+	SoupSocketPrivate *priv;
 	GIOStatus status;
+#ifdef SIGPIPE
 	gpointer pipe_handler;
+#endif
 	GIOCondition cond = G_IO_OUT;
 	GError *err = NULL;
 
 	g_return_val_if_fail (SOUP_IS_SOCKET (sock), SOUP_SOCKET_ERROR);
+	priv = SOUP_SOCKET_GET_PRIVATE (sock);
 
-	g_mutex_lock (sock->priv->iolock);
+	g_mutex_lock (priv->iolock);
 
-	if (!sock->priv->iochannel) {
-		g_mutex_unlock (sock->priv->iolock);
+	if (!priv->iochannel) {
+		g_mutex_unlock (priv->iolock);
 		return SOUP_SOCKET_EOF;
 	}
-	if (sock->priv->write_tag) {
-		g_mutex_unlock (sock->priv->iolock);
+	if (priv->write_src) {
+		g_mutex_unlock (priv->iolock);
 		return SOUP_SOCKET_WOULD_BLOCK;
 	}
 
+#ifdef SIGPIPE
 	pipe_handler = signal (SIGPIPE, SIG_IGN);
-	status = g_io_channel_write_chars (sock->priv->iochannel,
+#endif
+	status = g_io_channel_write_chars (priv->iochannel,
 					   buffer, len, nwrote, &err);
+#ifdef SIGPIPE
 	signal (SIGPIPE, pipe_handler);
+#endif
 	if (err) {
 		if (err->domain == SOUP_SSL_ERROR &&
 		    err->code == SOUP_SSL_ERROR_HANDSHAKE_NEEDS_READ)
@@ -1115,19 +1350,30 @@ soup_socket_write (SoupSocket *sock, gconstpointer buffer,
 				   NULL);
 	}
 
+	/* If the connection/session is sync and  we get
+	   EAGAIN or EWOULDBLOCK, then it will be socket timeout
+	   and should be treated as an error condition.
+	*/
+	if (!priv->non_blocking && status == G_IO_STATUS_AGAIN) {
+		g_mutex_unlock (priv->iolock);
+		return SOUP_SOCKET_ERROR;
+	}
+
 	if (status != G_IO_STATUS_NORMAL && status != G_IO_STATUS_AGAIN) {
-		g_mutex_unlock (sock->priv->iolock);
+		g_mutex_unlock (priv->iolock);
 		return SOUP_SOCKET_ERROR;
 	}
 
 	if (*nwrote) {
-		g_mutex_unlock (sock->priv->iolock);
+		g_mutex_unlock (priv->iolock);
 		return SOUP_SOCKET_OK;
 	}
 
-	sock->priv->write_tag =
-		g_io_add_watch (sock->priv->iochannel, cond, 
-				socket_write_watch, sock);
-	g_mutex_unlock (sock->priv->iolock);
+	priv->write_src =
+		soup_add_io_watch (priv->async_context,
+				   priv->iochannel,
+				   cond | G_IO_HUP | G_IO_ERR, 
+				   socket_write_watch, sock);
+	g_mutex_unlock (priv->iolock);
 	return SOUP_SOCKET_WOULD_BLOCK;
 }
